/* Copyright (C) 2019 Mono Wireless Inc. All Rights Reserved.    *
   Released under MW-SLA-*J,*E (MONO WIRELESS SOFTWARE LICENSE
   AGREEMENT).                                                   */

/*
   本コードはTWELITE標準シリアル電文形式のパーサの簡単な使用例です。
*/

#include <SoftwareSerial.h>
#include <sercmd.h>

String VersionNumber = "1.0.0";

// typedef
typedef unsigned short uint16;
typedef unsigned long uint32;

// 受信したデータを保管する構造体
typedef struct {
  bool ExPacket;      // 拡張形式か否か
  byte LID;           // 送信元の論理デバイスID
  byte RspID;         // 応答ID
  uint16 Length;      // データの長さ
  byte Data[255];     // データ
  /* 拡張形式でない場合は以下の変数は0xFFFFFFFFになる */
  uint32 DistExAddr;  // 送信元のシリアルID
  uint32 RcvExAddr;   // 送信先のシリアルID(論理デバイスIDを指定して送信している場合は0xFFFFFFFF)
  byte LQI;           // 通信品質(高いほど良い)
  /* ここまで */
} tsAppUart;

// Defines
#define S_OCTET(c) { *q = c; q++; }

#define G_OCTET(c) { c = *p; p++; }
#define G_BE_WORD(c) { c = (uint16)*(p+1) | ((uint16)*p << 8); p+=2; }
#define G_BE_DWORD(c) { c = (uint32)*(p+3) | ((uint32)*(p+2) << 8) | ((uint32)*(p+1) << 16) | ((uint32)*(p) << 24); p+=4; }

#define SW 8
#define LED 9

// Prototype
void vMWPutChar( byte chr );
bool bTWELITEDataArrived();
void vPrintHex( byte b );
tsAppUart tsParseAppUart( sercmd SerailData );
void vSendMessage( byte* OutputData, byte Length );

// Global Variables
SoftwareSerial MWSerial(2, 3); // RX, TX
sercmd SerialIn(true, 0, NULL );            // シリアル読み込み用のオブジェクト
sercmd SerialOut(true, 0, vMWPutChar );      // シリアル書き込み用のオブジェクト
byte u8MyAddr = 0x78;              // 自分の論理デバイスID
byte u8DistAddr = 0x00;            // 送信先の論理デバイスID


void setup() {
  Serial.begin(38400);
  Serial.print("MONOWIRELESS Serial Parser v");
  Serial.println(VersionNumber);

  pinMode( SW, INPUT );
  pinMode( LED, OUTPUT );
  digitalWrite( LED, LOW );

  // 内蔵のLEDを送れたかどうかのステータスにする
  pinMode( LED_BUILTIN, OUTPUT );
  digitalWrite( LED_BUILTIN, LOW );

  MWSerial.begin(38400);

  // TWELITEが親機かどうか確認するメッセージを送る
  byte cmd[] = {0xDB, 0xF1};
  // バイト列のコピー
  memcpy( SerialOut.au8data, cmd, 2 );
  // バイト列の長さを指定する
  SerialOut.u16len = 2;
  // TWELITEにコマンドを送信する (なぜか2回送らないと受信されない)
  SerialOut.vOutput();
  SerialOut.vOutput();
}

void loop() {
  static uint32 ct = 0;
  // 解釈に成功したら終了
  if (bTWELITEDataArrived()) {
    // 系列を読み出せたので構造体に代入する
    tsAppUart sAppUart = tsParseAppUart( SerialIn );

    // TWELITE から来たシリアルメッセージがDBコマンドだったら以下を処理する。
    if ( sAppUart.LID == 0xDB ) {
      switch (sAppUart.Data[0]) {
        case 0xA1:    // 送信したときの応答メッセージ
          if (sAppUart.Data[2]) {
            Serial.print("Transmission success");
            digitalWrite( LED_BUILTIN, HIGH );
            ct = 5000;
          } else {
            Serial.print("Transmission failure");
          }
          Serial.println();
          break;
        case 0xF1:    // 自分のアドレス群
          u8MyAddr = sAppUart.Data[9];
          Serial.print("MyAddr = 0x");
          vPrintHex(u8MyAddr);
          //  親機なら送信先を子機に変更する
          if ( u8MyAddr == 0x00 ) {
            u8DistAddr = 0x78;
          }
          break;
      }
    } else {
      // 系列を読み出せたので表示
      Serial.println("Data recieved.");
      Serial.print("Sender LID = 0x");
      vPrintHex(sAppUart.LID);
      Serial.print("Response ID = 0x");
      vPrintHex(sAppUart.RspID);
      Serial.print("Message = ");
      for (int i = 0; i < sAppUart.Length; i++) {
        Serial.print((char)(sAppUart.Data[i]));
      }
      Serial.println();

      // 自分の送信先と送られてきたパケットの送信元が一緒だったらLEDの状態を変化させる
      // 本スケッチを参考に修正する場合はここを修正する。
      if ( sAppUart.LID == u8DistAddr ) {
        if (sAppUart.Length == 1) {
          if ( sAppUart.Data[0] == 1 ) {
            digitalWrite( LED, HIGH );
            Serial.println("Trun On.");
          } else if (sAppUart.Data[0] == 0) {
            digitalWrite( LED, LOW );
            Serial.println("Trun Off.");
          }
        }
      }

      Serial.println();
    }
  }

  // タクトスイッチの状態を確認し、前の状態と変化していたら送信する
  static byte bSwState = 0x00;
  if ( bSwState != digitalRead(SW) ) {
    bSwState = !bSwState;
    // 無線送信をする
    vSendMessage( &bSwState, 1 );
  }

  // 一定時間たったら内蔵LEDを消す
  if (ct != 0) {
    ct--;
    if (ct == 0) {
      digitalWrite( LED_BUILTIN, LOW );
    }
  }
}

// 受信したパケットを解釈し、構造体に入れる関数
tsAppUart tsParseAppUart( sercmd SerialData )
{
  tsAppUart sAppUart;
  byte* p = SerialData.au8data;
  G_OCTET(sAppUart.LID);

  // DBコマンドだったらそのままDataに入れる
  if( sAppUart.LID == 0xDB ){
    sAppUart.Length = SerialData.u16len - 1;
  }else{
    // 拡張形式だったらフラグを立てる
    if (*p == 0xA0) {
      sAppUart.ExPacket = true;
      p++;
    } else {
      sAppUart.ExPacket = false;
    }


    G_OCTET(sAppUart.RspID);

    sAppUart.DistExAddr = 0xFFFFFFFF;
    sAppUart.RcvExAddr = 0xFFFFFFFF;
    sAppUart.LQI = 0xFF;
    sAppUart.Length = SerialData.u16len - 2;

    if ( sAppUart.ExPacket ) {
      G_BE_DWORD(sAppUart.DistExAddr);
      G_BE_DWORD(sAppUart.RcvExAddr);
      G_OCTET(sAppUart.LQI);
      G_BE_WORD(sAppUart.Length);
    }
  }
  for (int i = 0; i < sAppUart.Length; i++) {
    G_OCTET(sAppUart.Data[i]);
  }

  return sAppUart;
}

// MAC ACK付きの送信メッセージを作成しTWELITEに送る関数
void vSendMessage( byte* OutputData, byte Length )
{
  byte* q = SerialOut.au8data;
  S_OCTET(u8DistAddr);  // 送信先
  S_OCTET(0xA0);        // 拡張形式
  S_OCTET(0x01);        // 応答ID
  S_OCTET(0x01);        // MAC ACK
  S_OCTET(0xFF);        // オプションここまで

  // ペイロード
  byte MaxLength = (Length < 80) ? Length : 80;
  for(byte i = 0; i<MaxLength; i++){
    S_OCTET(OutputData[i]);
  }

  // バイト列の長さを指定する
  SerialOut.u16len = q - SerialOut.au8data;
  // TWELITEにコマンドを送信する
  SerialOut.vOutput();

}

void vMWPutChar( byte chr )
{
  MWSerial.write(chr);
}

bool bTWELITEDataArrived()
{
  // TWELITEからの出力を確認する。
  int c = MWSerial.read();
  if (c >= 0) {
    // １バイトずつ解釈をしていく
    byte u8stat = SerialIn.u8Parse(c);
    // 解釈に成功したら終了
    if (u8stat == E_SERCMD_COMPLETE) {
      return true;
    } else if (u8stat == E_SERCMD_CHECKSUM_ERROR) {
      // 電文中のチェックサムと計算したチェックサムが異なっているので、おそらく正しく読み込めなかった
      Serial.print("Command checksum error\n");
    }
  }
  return false;
}

void vPrintHex( byte b )
{
  if (b < 0x10) Serial.print(0);
  Serial.println(b, HEX);
}
