/* Copyright (C) 2017 Mono Wireless Inc. All Rights Reserved.    *
 * Released under MW-SLA-*J,*E (MONO WIRELESS SOFTWARE LICENSE   *
 * AGREEMENT).                                                   */

/** @file
 *
 * @defgroup SERCMD シリアルコマンドの解釈を行う
 * シリアルコマンドの解釈を行う。解釈部は書式に応じてアスキー形式、バイナリ形式、必要で
 * あるなら本定義に基づき拡張も可能である。
 *
 * - 入力の仕様（アスキー形式）
 *   - :([0-9A-F][0-9A-F])+[0-9A-F][0-9A-F][CR][LF]
 *   - 最後の 2 バイトは LRC となっていてLRCを含む 16進変換したバイトを加算すると 0 になる。\n
 *     つまり LRC は全データの和を取り、符号を反転（二の補数）させたもの。
 * - 入力の仕様（バイナリ形式）
 *   - 0xA5 0x5A [LenH] [LenL] [Payload] [XOR] [EOT]
 *      - LenH: ペイロードの長さ(バイト)の上位バイト (MSB を 0x80 にすること)
 *      - LenL: ペイロードの長さ(バイト)の下位バイト
 *      - Payload: ペイロード
 *      - XOR: ペイロード部の各バイトの XOR 値を用いたチェックサム
 *      - EOT: 省略可能。vOutput() 出力時には付与される。
 *
 * - API 利用方法
 *   - 準備
 *     - 事前に tsSerCmd_Context sSC 構造体および入力バッファデータ領域を確保しておく
 *     - SerCmdAscii_vInit() SerCmdBinary_vInit() 関数により構造体を初期化する
 *     - 入力タイムアウトを設定する場合は tsSerCmd_Context 構造体の u16timeout メンバーの値[ms]を設定する
 *
 *   - 解釈実行
 *     - UART からの入力バイト(u8Byte)を  uint8 u8ret = sSC.u8Parse(&sSC, u8Byte) に入力する
 *     - u8ret の値が 0x80 以上になると、解釈が終了したことを示す
 *        - E_SERCMD_COMPLETE: エラーなく系列が入力された。この呼び出しコンテキスト中で処理すること。
 *                             続く入力処理 (->u8Parse(), ->bComplete())が行われた時点でデータの一貫性
 *                             は保たれない。
 *        - E_SERCMD_ERROR: 解釈エラー
 *        - E_SERCMD_CHECKSUM_ERROR: チェックサムだけ合わなかった。このとき sSC.u16cksum に期待したチェックサム値が格納される。
 *     - 続けて ->u8Parse() を実行する。直前の状態が終了・エラー状態であっても、新たな系列として解釈を始めます。
 *
 *   - タイムアウトの処理
 *     外部のタイマから関数 ->bComplete() を実行するとタイムアウトのチェックを行う。
 *     - bComplete() が実装されていない(=NULL)の場合はタイムアウト処理は行えない。
 *     - 関数 ->bComplete() が TRUE を戻した場合は、呼び出しコンテキスト中でデータを利用すること。
 *       続く入力処理 (->u8Parse(), ->bComplete())が行われた時点でデータの一貫性は保たれない。
 *
 */

#ifndef SERCMD_H_
#define SERCMD_H_

#define MAX_LEN 255

/** @ingroup SERCMD
 * シリアルコマンド解釈の内部状態。
 * - 本状態は各解釈部共通の定義とする
 */
typedef enum {
	E_SERCMD_EMPTY = 0,                //!< 入力されていない
	E_SERCMD_COMPLETE = 0x80,          //!< 入力が完結した(LCRチェックを含め)
	E_SERCMD_ERROR = 0x81,             //!< 入力エラー
	E_SERCMD_CHECKSUM_ERROR = 0x82,   //!< チェックサムが間違っている
} teSerCmdGenState;

class sercmd
{
private:
	unsigned short u16pos; //!< 入力位置（内部処理用）
	unsigned short  u16cksum; //!< チェックサム
	bool bBinary;	//!< TrueだったらBinary書式、FalseだったらAscii書式
	unsigned char u8state; //!< 状態

	void *vExtraData; //!< 処理系独特の追加データ

	void (*vPutChar)(unsigned char u8byte); //!< 出力のコールバック関数

	unsigned char u8BinaryParse(unsigned char u8byte);
	void vBinaryOutput();
	unsigned char u8AsciiParse(unsigned char u8byte);
	void vPutByte(unsigned char u8byte, unsigned char *pu8lrc);
	void vAsciiOutput();

public:
	unsigned short u16len; //!< ペイロード長
	unsigned char* au8data; //!< バッファ（初期化時に固定配列へのポインタを格納しておく)
	unsigned short u16maxlen; //!< バッファの最大長

	unsigned long u32timestamp; //!< タイムアウトを管理するためのカウンタ
	unsigned short u16timeout; //!< 入力系列のタイムアウト (0ならタイムアウト無し)

	sercmd();
	sercmd( bool bbin, unsigned short timeout, void (*cbPutChar)(unsigned char u8byte) );
	~sercmd();
	unsigned char u8Parse( unsigned char u8byte );
	void vOutput();
};

#endif /* SERCMD_GEN_H_ */
